/**************************************************************************
   findwild      general wildcard file and string search

   Copyright 2007, 2008, 2009, 2010  Michael Cornelison
   source URL:  kornelix.squarespace.com
   contact: kornelix@yahoo.de
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "zfuncs.h"

#define findwild_title "findwild v.1.1"
#define cbmem 20                                                  //  combobox depth
#define cbcc 200                                                  //  combobox max cc
#define Smax 10                                                   //  max. search/ignore string count
#define normfont "monospace 9"                                    //  report fonts
#define boldfont "monospace bold 9"
                                                                  //  file matching and ignoring rules:
#define match_any          1                                      //  find files with any search string
#define match_all          2                                      //  find files with all search strings
#define match_rec_all      3                                      //  find files with all search strings in one record
#define ignore_any         1                                      //  ignore files with any ignore string
#define ignore_all         2                                      //  ignore files with all ignore strings
#define ignore_rec_all     3                                      //  ignore files with all ignore strings in one record
#define ignore_match_any   4                                      //  ignore match record with any ignore string
#define ignore_match_all   5                                      //  ignore match record with all ignore strings

PangoFontDescription    *font;
GtkWidget      *mWin, *mVbox, *mScroll, *mLog;                             //  main window widgets
GtkTextBuffer  *textBuff;
GtkWidget      *toolbar, *stbar;

int initfunc(void * data);                                                 //  gtk_main() initial function

void buttfunc(GtkWidget *item, const char *menu);                          //  toolbar button response functions
void m_search();
void m_save();
void m_print();
void m_kill();
void m_clear();
void m_help();

int search_dialog_event(zdialog *zd, const char *event);                   //  widget event response function
int search_dialog_compl(zdialog *zd, int zstat);                           //  dialog completion function
void * filesearch_thread(void *);                                          //  directory scan / file search thread
void break_criteria(char *string, char *strings[Smax], int &count);        //  break search/ignore strings in substrings
int filesearch(const char *file);                                          //  search file for matching string

int load_criteria();                                                       //  history file >> search criteria
int save_criteria();                                                       //  search criteria >> history file
int search_dialog_stuff(zdialog *zd);                                      //  search criteria >> dialog widgets
int search_dialog_fetch(zdialog *zd);                                      //  dialog widgets >> search criteria

char        *sr_path[cbmem], *sr_file[cbmem], *sr_string[cbmem];           //  search criteria from dialog
char        *ig_file[cbmem], *ig_string[cbmem];
time_t      dt_from, dt_to;
int         matchrule, ignorerule;                                         //  match rule and ignore rule from dialog

const char  *mstext[3] = { "any search string","all search strings",
                           "all search strings in same record" };

const char  *igtext[5] = { "any ignore string","all ignore strings",
                           "all ignore strings in same record",
                           "with any ignore string","with all ignore strings" };

char        *srfiles[Smax], *srstrings[Smax];                              //  0-Smax search files and strings
char        *igfiles[Smax], *igstrings[Smax];                              //  0-Smax ignore files and strings
int         nsrf, nsrs, nigf, nigs;                                        //  actual counts

char        *delims[cbmem];                                                //  string delimiters
const char  defaultdelims[] = " =()[]{}.,;:'<>!-+*/|~`%^&?\\\"";

int         dialogbusy = 0, killsearch = 0, searchbusy = 0;                //  flags
int         listmatch = 0;

char        criteriaFile[1000];                                            //  file - save search criteria
char        hitsFile[1000];                                                //  file - save search hits (found files)
char        hitsFile2[1000];                                               //  file - read search hits
char        workbuff[1000];
char        workbuff2[1000];


//  main windowing program

int main(int argc, char *argv[])
{
   gtk_init(&argc,&argv);                                                  //  GTK command line options
   if (! g_thread_supported())                                             //  suddenly required for new gtk   v.05
         g_thread_init(null);                                              //  initz. GTK for threads
   gdk_threads_init();
   zlockInit();                                                            //  initz. GTK/GDK for threads

   initz_appfiles("findwild",null);                                        //  set up app directories

   mWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);                             //  main window
   gtk_window_set_title(GTK_WINDOW(mWin),findwild_title);
   gtk_window_set_position(GTK_WINDOW(mWin),GTK_WIN_POS_CENTER);
   gtk_window_set_default_size(GTK_WINDOW(mWin),800,400);
   
   mVbox = gtk_vbox_new(0,0);                                              //  vertical packing box
   gtk_container_add(GTK_CONTAINER(mWin),mVbox);                           //  add to main window

   toolbar = create_toolbar(mVbox);                                        //  create tool bar and buttons

   add_toolbar_button(toolbar,"search","start search","gtk-find",buttfunc);
   add_toolbar_button(toolbar,"save","save output","gtk-save",buttfunc);
   add_toolbar_button(toolbar,"print","print output","gtk-print",buttfunc);
   add_toolbar_button(toolbar,"kill","kill search","gtk-stop",buttfunc);
   add_toolbar_button(toolbar,"clear","clear output","gtk-clear",buttfunc);
   add_toolbar_button(toolbar,"quit","quit findwild","gtk-quit",buttfunc);
   add_toolbar_button(toolbar,"help","show user guide","gtk-help",buttfunc);

   mScroll = gtk_scrolled_window_new(0,0);                                 //  scrolled window
   gtk_box_pack_start(GTK_BOX(mVbox),mScroll,1,1,0);                       //  add to main window mVbox
   mLog = gtk_text_view_new();                                             //  text window
   gtk_container_add(GTK_CONTAINER(mScroll),mLog);                         //  add to scrolled window
   textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(mLog));               //  get related text buffer

   stbar = create_stbar(mVbox);                                            //  add status bar

   font = pango_font_description_from_string(normfont);                    //  set window font
   gtk_widget_modify_font(mLog,font);

   gtk_widget_show_all(mWin);                                              //  show all widgets
   
   G_SIGNAL(mWin,"destroy",gtk_main_quit,0)                                //  connect window destroy event

   gtk_init_add((GtkFunction) initfunc,0);                                 //  setup initial call from gtk_main()

   gdk_threads_enter();
   gtk_main();                                                             //  process events
   gdk_threads_leave();

   return 0;
}


//  initial function called from gtk_main() at startup

int initfunc(void * data)
{
   int      ii;

   for (ii = 0; ii < cbmem; ii++)                                          //  clear memory lists
   {
      sr_path[ii] = 0;
      sr_file[ii] = 0;
      sr_string[ii] = 0;
      ig_file[ii] = 0;
      ig_string[ii] = 0;
      delims[ii] = 0;
   }
   
   for (ii = 0; ii < Smax; ii++)
   {
      srfiles[ii] = 0;
      srstrings[ii] = 0;
      igfiles[ii] = 0;
      igstrings[ii] = 0;
   }
   
   dt_from = dt_to = 0;

   *hitsFile = 0;                                                          //  set up search hits save file
   strncatv(hitsFile,999,get_zuserdir(),"/search_hits",null);

   *criteriaFile = 0;                                                      //  set up search criteria save file
   strncatv(criteriaFile,999,get_zuserdir(),"/search_criteria",null);
   load_criteria();                                                        //  get last saved search criteria

   return 0;
}


//  toolbar button response functions

void buttfunc(GtkWidget *item, const char *menu)
{
   if (strstr(menu,"search")) m_search();
   if (strstr(menu,"save")) m_save();
   if (strstr(menu,"print")) m_print();
   if (strstr(menu,"kill")) m_kill();
   if (strstr(menu,"clear")) m_clear();
   if (strstr(menu,"help")) m_help();
   if (strstr(menu,"quit")) gtk_main_quit();
}


//  save screen to file

void m_save()
{
   wfilesave(mLog);
   return;
}


//  print screen

void m_print()
{
   wprintp(mLog);
   return;
}


//  kill running search

void m_kill()
{
   wprintf(mLog,"kill ... \n");

   killsearch = 1;

   while (searchbusy) {
      zmainloop();
      zsleep(0.1);
   }

   killsearch = 0;
   return;
}


//  clear screen

void m_clear()
{
   wprintf(mLog,"clear \n");
   zsleep(0.1);
   gtk_text_buffer_set_text(textBuff,"", -1);
   return;
}


//  launch help file in new process

void m_help()
{
   showz_userguide();
   return;
}


//  build and initiate search dialog

void m_search()
{
   if (dialogbusy) return;
   dialogbusy = 1;

   /***********************************************************************

      Find files containing:     Ignore files containing:
       [x] any search string      [x] any ignore string
       [_] all search strings     [_] all ignore strings
       [_] all in same record     [_] all in same record

                                 Ignore matching records with: 
                                  [_] any ignore string
                                  [_] all ignore strings

              search path  [_________________________][V]  [clear]
           search file(s)  [_________________________][V]  [clear]
         search string(s)  [_________________________][V]  [clear]
           ignore file(s)  [_________________________][V]  [clear]
         ignore string(s)  [_________________________][V]  [clear]
        string delimiters  [_________________________][V]  [clear]

       date from  [_______]   to [_______]  -days or yyyy-mm-dd

       [x] list match recs  [search all] [search hits] [save criteria]

   ************************************************************************/

   zdialog     *zd;
   zd = zdialog_new("findwild - search for files",mWin,"cancel",null);

   zdialog_add_widget(zd,"hbox","hbr1","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","space","hbr1",0,"space=10");
   zdialog_add_widget(zd,"vbox","vbr1","hbr1");
   zdialog_add_widget(zd,"label","rules1","vbr1","Find files containing:  ");
   zdialog_add_widget(zd,"radio","ruleM1","vbr1","any search string");
   zdialog_add_widget(zd,"radio","ruleM2","vbr1","all search strings");
   zdialog_add_widget(zd,"radio","ruleM3","vbr1","all in same record");
   zdialog_add_widget(zd,"vbox","space","hbr1",0,"space=10");
   zdialog_add_widget(zd,"vbox","vbr2","hbr1");
   zdialog_add_widget(zd,"hbox","hbr2","vbr2");
   zdialog_add_widget(zd,"label","rules2","hbr2","Ignore files containing:");
   zdialog_add_widget(zd,"radio","ruleI1","vbr2","any ignore string");
   zdialog_add_widget(zd,"radio","ruleI2","vbr2","all ignore strings");
   zdialog_add_widget(zd,"radio","ruleI3","vbr2","all in same record");
   zdialog_add_widget(zd,"hbox","space","vbr2",0,"space=3");
   zdialog_add_widget(zd,"label","rules3","vbr2","Ignore matching records with:");
   zdialog_add_widget(zd,"radio","ruleI4","vbr2","any ignore strings");
   zdialog_add_widget(zd,"radio","ruleI5","vbr2","all ignore strings");

   zdialog_add_widget(zd,"hbox","hbs1","dialog",0,"space=10");
   zdialog_add_widget(zd,"vbox","vbs11","hbs1",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vbs12","hbs1",0,"homog|expand|space=5");
   zdialog_add_widget(zd,"vbox","vbs13","hbs1",0,"homog|space=5");
   zdialog_add_widget(zd,"label","lab_sr_path","vbs11","      search path");
   zdialog_add_widget(zd,"label","lab_sr_file","vbs11","     search file(s)");
   zdialog_add_widget(zd,"label","lab_sr_string","vbs11"," search string(s)");
   zdialog_add_widget(zd,"label","lab_ig_file","vbs11","     ignore file(s)");
   zdialog_add_widget(zd,"label","lab_ig_string","vbs11"," ignore string(s)");
   zdialog_add_widget(zd,"label","lab_delims","vbs11","string delimiters");
   zdialog_add_widget(zd,"comboE","sr_path","vbs12",0,"expand");
   zdialog_add_widget(zd,"comboE","sr_file","vbs12",0,"expand"); 
   zdialog_add_widget(zd,"comboE","sr_string","vbs12",0,"expand");
   zdialog_add_widget(zd,"comboE","ig_file","vbs12",0,"expand");
   zdialog_add_widget(zd,"comboE","ig_string","vbs12",0,"expand");
   zdialog_add_widget(zd,"comboE","delims","vbs12",0,"expand");
   zdialog_add_widget(zd,"button","clear_sr_path","vbs13","clear");
   zdialog_add_widget(zd,"button","clear_sr_file","vbs13","clear");
   zdialog_add_widget(zd,"button","clear_sr_string","vbs13","clear");
   zdialog_add_widget(zd,"button","clear_ig_file","vbs13","clear");
   zdialog_add_widget(zd,"button","clear_ig_string","vbs13","clear");
   zdialog_add_widget(zd,"button","clear_delims","vbs13","clear");

   zdialog_add_widget(zd,"hbox","hbd2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab_dt_from","hbd2","  date from");
   zdialog_add_widget(zd,"entry","dt_from","hbd2","-999","scc=11");
   zdialog_add_widget(zd,"label","lab_dt_to","hbd2","  to");
   zdialog_add_widget(zd,"entry","dt_to","hbd2","0","scc=10");
   zdialog_add_widget(zd,"label","x","hbd2","-days or yyyy-mm-dd","space=10");

   zdialog_add_widget(zd,"hbox","hbt1","dialog",0,"space=10");
   zdialog_add_widget(zd,"check","list match","hbt1","list matching records","space=10");
   zdialog_add_widget(zd,"button","search all","hbt1","search all");
   zdialog_add_widget(zd,"button","search hits","hbt1","search hits");
   zdialog_add_widget(zd,"button","save criteria","hbt1","save criteria");

   zdialog_resize(zd,500,0);
   search_dialog_stuff(zd);                                                //  search criteria >> dialog_widgets
   zdialog_run(zd,search_dialog_event,search_dialog_compl);                //  start dialog (non modal)
   return;
}


//  search dialog - widget event response function

int search_dialog_event(zdialog *zd, const char *event)
{
   static int  fhits;
   
   if (strEqu(event,"clear_sr_path")) zdialog_stuff(zd,"sr_path","");      //  clear buttons
   if (strEqu(event,"clear_sr_file")) zdialog_stuff(zd,"sr_file","");
   if (strEqu(event,"clear_sr_string")) zdialog_stuff(zd,"sr_string","");
   if (strEqu(event,"clear_ig_file")) zdialog_stuff(zd,"ig_file","");
   if (strEqu(event,"clear_ig_string")) zdialog_stuff(zd,"ig_string","");
   if (strEqu(event,"clear_delims")) zdialog_stuff(zd,"delims","");

   if (strEqu(event,"search all") || strEqu(event,"search hits"))
   {                                                                       //  search all button
      if (searchbusy) {
         zmessageACK("wait for prior search to complete");
         return 0;
      }
      
      fhits = 0;
      if (strEqu(event,"search hits")) fhits = 1;
      
      search_dialog_fetch(zd);                                             //  dialog widgets >> search criteria
      start_detached_thread(filesearch_thread,&fhits);                     //  start search thread
      searchbusy = 1;
   }

   if (strEqu(event,"save criteria"))                                      //  save search criteria in a file
   {
      wprintf(mLog,"save search criteria \n");
      search_dialog_fetch(zd);                                             //  dialog widgets >> search criteria
      save_criteria();                                                     //  search criteria >> history file
   }

   return 0;
}


//  search dialog - completion function

int search_dialog_compl(zdialog *zd, int zstat)
{
   zdialog_destroy(zd);
   dialogbusy = 0;
   return 0;
}


//  break search criteria into 0-Smax search/ignore substrings

void break_criteria(char *string, char *strings[Smax], int &count)
{
   int      ii;
   cchar     *pp;

   count = 0;
   if (! strlen(string)) return;

   for (ii = 0; ii < Smax; ii++)
   {
      if (strings[ii]) zfree(strings[ii]);
      strings[ii] = 0;
   }

   for (ii = 0; ii < Smax; ii++)
   {
      pp = strField(string,' ',ii+1);
      if (! pp) break;
      strings[ii] = strdupz(pp);
   }

   count = ii;
   return;
}


//  load last saved search criteria from file

int load_criteria()
{
   int  load_criteria2(FILE *fid, char **memlist, int count);
   FILE        *fid = null;
   char        *pp, criteria[20];
   int         count, nf;
   
   fid = fopen(criteriaFile,"r");                                          //  criteria file
   if (! fid) return 0;

   while (true)
   {
      pp = fgets_trim(workbuff,20,fid,1);                                  //  read criteria name, record count
      if (! pp) break;
      nf = sscanf(workbuff,"%s %d",criteria,&count);
      if (nf != 2) break;
      if (count < 0 || count > cbmem) break;
      if (strEqu(criteria,"searchpath")) load_criteria2(fid,sr_path,count);
      if (strEqu(criteria,"searchfile")) load_criteria2(fid,sr_file,count);
      if (strEqu(criteria,"searchstring")) load_criteria2(fid,sr_string,count);
      if (strEqu(criteria,"ignorefile")) load_criteria2(fid,ig_file,count);
      if (strEqu(criteria,"ignorestring")) load_criteria2(fid,ig_string,count);
      if (strEqu(criteria,"stringdelims")) load_criteria2(fid,delims,count);
   }

   fclose(fid);
   return 1;
}

int load_criteria2(FILE *fid, char **memlist, int count)                   //  helper function
{
   int      ii;
   char     *pp;

   for (ii = 0; ii < count; ii++)                                          //  load memory list from file recs
   {
      pp = fgets_trim(workbuff,999,fid,1);
      if (! pp) return ii;
      if (memlist[ii]) zfree(memlist[ii]);
      memlist[ii] = strdupz(workbuff);
   }

   return ii;
}


//  save search criteria in memory >> findwild history file

int save_criteria()
{
   int  save_criteria2(FILE *fid, cchar *criteria, char **memlist);
   FILE        *fid = null;

   fid = fopen(criteriaFile,"w");
   if (! fid) zappcrash("cannot open search_criteria output file");
   save_criteria2(fid,"searchpath",sr_path);
   save_criteria2(fid,"searchfile",sr_file);
   save_criteria2(fid,"searchstring",sr_string);
   save_criteria2(fid,"ignorefile",ig_file);
   save_criteria2(fid,"ignorestring",ig_string);
   save_criteria2(fid,"stringdelims",delims);
   fclose(fid);
   return 0;
}

int save_criteria2(FILE *fid, cchar *criteria, char **memlist)             //  helper function
{
   int      ii, nf;

   for (nf = 0; nf < cbmem; nf++)                                          //  count combo list entries
         if (memlist[nf] == 0) break;
   fprintf(fid,"%s %d""\n",criteria,nf);                                   //  output criteria name, count
   for (ii = 0; ii < nf; ii++) 
         fprintf(fid,"%s""\n",memlist[ii]);                                //  output criteria list entries
   return nf;
}


//  search criteria lists in memory >> dialog widgets

int search_dialog_stuff(zdialog *zd)
{
   zdialog_stuff(zd,"ruleM1",1);                                           //  default match and ignore rules
   zdialog_stuff(zd,"ruleI1",1);

   if (sr_path[0]) zdialog_stuff(zd,"sr_path",sr_path[0]);                 //  stuff 1st value in edit boxes
   if (sr_file[0]) zdialog_stuff(zd,"sr_file",sr_file[0]);
   if (sr_string[0]) zdialog_stuff(zd,"sr_string",sr_string[0]);
   if (ig_file[0]) zdialog_stuff(zd,"ig_file",ig_file[0]);
   if (ig_string[0]) zdialog_stuff(zd,"ig_string",ig_string[0]);
   if (delims[0]) zdialog_stuff(zd,"delims",delims[0]);
   else zdialog_stuff(zd,"delims",defaultdelims);

   for (int ii = 0; ii < cbmem; ii++)                                      //  append all values to combo box
   {                                                                       //    drop-down lists 
      if (sr_path[ii]) zdialog_cb_app(zd,"sr_path",sr_path[ii]);
      if (sr_file[ii]) zdialog_cb_app(zd,"sr_file",sr_file[ii]);
      if (sr_string[ii]) zdialog_cb_app(zd,"sr_string",sr_string[ii]);
      if (ig_file[ii]) zdialog_cb_app(zd,"ig_file",ig_file[ii]);
      if (ig_string[ii]) zdialog_cb_app(zd,"ig_string",ig_string[ii]);
      if (delims[ii]) zdialog_cb_app(zd,"delims",delims[ii]);
   }
   
   return 0;
}


//  dialog widgets >> search criteria lists in memory

int search_dialog_fetch(zdialog *zd)
{
   int  search_dialog_fetch2(zdialog *zd, cchar *wname, char **memlist);
   time_t  search_dialog_fetchdate(zdialog *zd, cchar *dt);
   int      rule;
   
   zdialog_fetch(zd,"ruleM1",rule); if (rule) matchrule = 1;               //  get match and ignore rules
   zdialog_fetch(zd,"ruleM2",rule); if (rule) matchrule = 2;
   zdialog_fetch(zd,"ruleM3",rule); if (rule) matchrule = 3;
   zdialog_fetch(zd,"ruleI1",rule); if (rule) ignorerule = 1;
   zdialog_fetch(zd,"ruleI2",rule); if (rule) ignorerule = 2;
   zdialog_fetch(zd,"ruleI3",rule); if (rule) ignorerule = 3;
   zdialog_fetch(zd,"ruleI4",rule); if (rule) ignorerule = 4;
   zdialog_fetch(zd,"ruleI5",rule); if (rule) ignorerule = 5;

   search_dialog_fetch2(zd,"sr_path",sr_path);                             //  get combobox entries
   search_dialog_fetch2(zd,"sr_file",sr_file);
   search_dialog_fetch2(zd,"sr_string",sr_string);
   search_dialog_fetch2(zd,"ig_file",ig_file);
   search_dialog_fetch2(zd,"ig_string",ig_string);
   search_dialog_fetch2(zd,"delims",delims);

   zdialog_fetch(zd,"list match",listmatch);                               //  list matching records, yes/no
   
   dt_from = search_dialog_fetchdate(zd,"dt_from");                        //  get date range
   dt_to = search_dialog_fetchdate(zd,"dt_to");

   return 0;
}


//  combobox entry and drop-down list >> memory table

int search_dialog_fetch2(zdialog *zd, cchar *wname, char **memlist)
{
   int         ii;
   char        *pp, cbent[cbcc];
   
   for (ii = 0; ii < cbmem; ii++)
   {
      if (memlist[ii]) zfree(memlist[ii]);                                 //  clear memlist
      memlist[ii] = 0;
   }
   
   zdialog_fetch(zd,wname,cbent,cbcc-1);                                   //  memlist[0] << combo entry
   memlist[0] = strdupz(cbent);

   for (ii = 1; ii < cbmem; ii++)
   {
      pp = zdialog_cb_get(zd,wname,ii-1);                                  //  memlist[1...] << combo dropdown list
      if (! pp) break;
      memlist[ii] = strdupz(pp);
   }

   return ii;
}


//  get a date from search dialog, format: -days  or  yyyy-mm-dd

time_t search_dialog_fetchdate(zdialog *zd, cchar *entry)
{
   char        temp[20];
   cchar       *pp;
   const char  delims[] = "-./ ";
   int         err, days, day, month, year;
   time_t      now1;                                                       //  numeric time in seconds since base
   struct tm   now2;                                                       //  year/month/day/hour/min/sec

   now1 = time(0);                                                         //  get current date/time in both formats
   now2 = *localtime(&now1);

   zdialog_fetch(zd,entry,temp,19);                                        //  user date entry
   pp = strField(temp,' ',1);
   if (! pp) return 0;

   err = convSI(pp,days,-999,0);                                           //  look for -999 to 0 days ago
   if (err == 0) return now1 + days*24*3600;                               //  OK, return NOW - seconds ago
   
   pp = strField(temp,delims,1);                                           //  look for yyyy-mm-dd format
   if (! pp) return 0;
   err = convSI(pp,year,1970,now2.tm_year+1900);                           //  year, 1970 to current
   if (err) return 0;
   pp = strField(temp,delims,2);
   if (! pp) return 0;
   err = convSI(pp,month,1,12);                                            //  month, 1-12
   if (err) return 0;
   pp = strField(temp,delims,3);
   if (! pp) return 0;
   err = convSI(pp,day,1,31);                                              //  day, 1-31
   if (err) return 0;
   now2.tm_year = year-1900;
   now2.tm_mon = month-1;
   now2.tm_mday = day;
   now1 = mktime(&now2);                                                   //  conv. to numeric time
   if (now1 < 0) return 0;
   return now1;
}


//  directory scan / file search function in separate thread

void * filesearch_thread(void *vhits)
{
   int            fhits, ccp, ccf, ii, jj;
   int            fcount, lcount, ftf, err;
   char           ch;
   const char     *pfile, *pname;
   struct tm      dfrom, dto;
   struct stat    statf;
   FILE           *fid = null, *fid2;
   
   fhits = *((int *) vhits);
   
   for (ii = 0; ii < (int) strlen(sr_string[0]); ii++)
   {
      ch = sr_string[0][ii];
      if (ch != ' ' && ch != '*' && strchr(delims[0],ch)) {
         zmessageACK("delimiter  %c  is contained in search string",ch);
         goto search_quit;
      }
   }

   for (ii = 0; ii < (int) strlen(ig_string[0]); ii++)
   {
      ch = ig_string[0][ii];
      if (ch != ' ' && ch != '*' && strchr(delims[0],ch)) {
         zmessageACK("delimiter  %c  is contained in ignore string",ch);
         goto search_quit;
      }
   }

   wprintf(mLog,"\n""begin search --------------------------- \n");        //  report all criteria used for search

   if (fhits) wprintf(mLog," search hits (files from previous search results) \n");
   wprintf(mLog,"   find files with: %s \n",mstext[matchrule-1]);
   if (ignorerule < 4) wprintf(mLog," ignore files with: %s \n",igtext[ignorerule-1]);
   else wprintf(mLog," ignore match recs: %s \n",igtext[ignorerule-1]);

   wprintf(mLog,"       search path: %s \n",sr_path[0]);
   wprintf(mLog,"       search file: %s \n",sr_file[0]);
   wprintf(mLog,"     search string: %s \n",sr_string[0]);
   wprintf(mLog,"    ignore file(s): %s \n",ig_file[0]);
   wprintf(mLog,"  ignore string(s): %s \n",ig_string[0]);
   wprintf(mLog," string delimiters: %s \n",delims[0]);

   if (dt_from || dt_to) {                                                 //  report date range if defined
      dfrom = *localtime(&dt_from);
      dto = *localtime(&dt_to);
      wprintf(mLog,"     mod date from: %d-%02d-%02d  to: %d-%02d-%02d \n",
                     dfrom.tm_year+1900, dfrom.tm_mon+1, dfrom.tm_mday,
                     dto.tm_year+1900, dto.tm_mon+1, dto.tm_mday);
   }
   
   wprintf(mLog,"\n");

   ccp = strlen(sr_path[0]);
   ccf = strlen(sr_file[0]);
   
   fcount = 0;

   if (! ccp) goto search_exit;                                            //  sanity checks
   if (! ccf) { sr_file[0] = strdupz("*"); ccf = 1; }                      //  bugfix           v.08
   if (ccp + ccf > 998) goto search_exit;
   if (dt_from > dt_to) goto search_exit;
   if (dt_from > time(0)) goto search_exit;

   break_criteria(sr_file[0],srfiles,nsrf);                                //  break search criteria into
   break_criteria(sr_string[0],srstrings,nsrs);                            //    search/ignore substrings
   break_criteria(ig_file[0],igfiles,nigf);
   break_criteria(ig_string[0],igstrings,nigs);
   
   if (! fhits)                                                            //  normal search
   {
      fid = fopen(hitsFile,"w");                                           //  open output file for search hits
      if (! fid) zappcrash("cannot open search_hits output file");

      for (ii = 0; ii < nsrf; ii++)                                        //  loop all search files
      {
         strcpy(workbuff,sr_path[0]);                                      //  combine pathname/filename to search
         if (sr_path[0][ccp-1] == '*' && srfiles[ii][0] == '*') 
               strcat(workbuff,srfiles[ii]+1);                             //  avoid path/**file
         else  strcat(workbuff,srfiles[ii]);

         ftf = 1;                                                          //  new SearchWild search

         while (true)
         {
            if (killsearch) {
               ftf = 2;                                                    //  abort search
               SearchWild(workbuff,ftf);
               break;
            }

            pfile = SearchWild(workbuff,ftf);                              //  find all matching files
            if (! pfile) break;

            pname = strrchr(pfile,'/') + 1;                                //  file name part
            stbar_message(stbar,pfile);                                    //  progress tracking in status bar

            for (jj = 0; jj < nigf; jj++)
            {
               if (MatchWild(igfiles[jj],pname) == 0) break;               //  file name part matches ignore file
               if (MatchWild(igfiles[jj],pfile) == 0) break;               //  whole filespec matches ignore file
            }
            if (jj < nigf) continue;                                       //  ignore file

            if (dt_to > 0) {                                               //  check mod date against range
               err = stat(pfile,&statf);
               if (err) continue;
               if (statf.st_mtime < dt_from ||                             //  out of range, ignore file
                   statf.st_mtime > dt_to) continue;
            }

            lcount = filesearch(pfile);                                    //  search for match strings, list hits
            if (lcount) {
               fprintf(fid,"%s""\n",pfile);                                //  write matching file to hits list
               fcount++;
            }
         }
      }

      fclose(fid);
   }
   
   if (fhits)                                                              //  search hits (previous files found)
   {
      strcpy(hitsFile2,hitsFile);                                          //  make copy of previous hits file
      strcat(hitsFile2,"_2");
      snprintf(workbuff,999,"mv -f %s %s",hitsFile,hitsFile2);
      err = system(workbuff);
      if (err) {
         zmessageACK("no previous files to search");
         goto search_quit;
      }

      fid2 = fopen(hitsFile2,"r");                                         //  open copy for reading previous hits
      if (! fid2) zappcrash("cannot open search_hits input file");
      
      fid = fopen(hitsFile,"w");                                           //  open hits for output of new hits
      if (! fid) zappcrash("cannot open search_hits output file");

      while (true)
      {
         if (killsearch) break;
         
         pfile = fgets_trim(workbuff,999,fid2);                            //  get next file from hit list
         if (! pfile) break;

         pname = strrchr(pfile,'/') + 1;                                   //  file name part
         stbar_message(stbar,pfile);                                       //  progress tracking in status bar
         
         for (ii = 0; ii < nsrf; ii++)                                     //  check for match with search files
            if (MatchWild(srfiles[ii],pname) == 0) break;                  //  (match file name only)
         if (ii == nsrf) continue;                                         //  no match

         for (jj = 0; jj < nigf; jj++)
         {
            if (MatchWild(igfiles[jj],pname) == 0) break;                  //  file name part matches ignore file
            if (MatchWild(igfiles[jj],pfile) == 0) break;                  //  whole filespec matches ignore file
         }
         if (jj < nigf) continue;                                          //  ignore file

         if (dt_to > 0) {                                                  //  check mod date against range
            err = stat(pfile,&statf);
            if (err) continue;
            if (statf.st_mtime < dt_from ||                                //  out of range, ignore file
                statf.st_mtime > dt_to) continue;
         }

         lcount = filesearch(pfile);                                       //  search for match strings, list hits
         if (lcount) {
            fprintf(fid,"%s""\n",pfile);                                   //  write matching file to hits list
            fcount++;
         }
      }

      fclose(fid);
      fclose(fid2);
   }
   
search_exit:
   if (killsearch) wprintf(mLog," *** search killed *** \n");
   else {
      wprintf(mLog,"\n %d files found \n",fcount);
      wprintf(mLog,"search completed ----------------------- \n");
   }

search_quit:
   searchbusy = 0;
   return 0;
}


//  file search function - match all file tokens to search string(s)

int filesearch(const char *filename)
{
   int      Fmatch[Smax], Fignore[Smax];                                   //  search and ignore strings in file
   int      Rmatch[Smax], Rignore[Smax];                                   //  search and ignore strings in record
   int      filematch, recmatch, recignore;
   FILE     *fid = null;
   char     buff1[1000], buff2[1000];
   char     *token, *prec;
   int      ii;
   
   if (nsrs == 0 && nigs == 0) {                                           //  no search or ignore strings (matches)
      wprintf(mLog," %s \n",filename);                                     //  output file name with no record counts
      return 1;
   }

   fid = fopen(filename,"r");                                              //  open file
   if (! fid) return 0;

   for (ii = 0; ii < Smax; ii++) Fmatch[ii] = Fignore[ii] = 0;             //  no strings found in file yet
   filematch = 0;

   while (true)
   {
      prec = fgets_trim(buff1,999,fid,1);                                  //  read next record
      if (! prec) break;

      for (ii = 0; ii < Smax; ii++) Rmatch[ii] = Rignore[ii] = 0;          //  no strings found in record yet
      recmatch = recignore = 0;

      while (true)
      {
         token = strtok(prec,delims[0]);                                   //  get next string defined by delimiters
         if (! token) break;
         prec = 0;

         for (ii = 0; ii < nsrs; ii++)
         {
            if (MatchWild(srstrings[ii],token) == 0) {                     //  match with all search strings
               Rmatch[ii]++;
               if (Rmatch[ii] == 1) recmatch++;                            //  search strings found, 0...nsrs
            }
         }

         for (ii = 0; ii < nigs; ii++)
         {
            if (MatchWild(igstrings[ii],token) == 0) {                     //  match with all ignore strings
               Rignore[ii]++;
               if (Rignore[ii] == 1) recignore++;                          //  ignore strings found, 0...nigs
            }
         }
      }
      
      if (recmatch + recignore == 0) continue;                             //  record has no matches, ignore

      if (nigs > 0) {
         if (ignorerule == ignore_any && recignore > 0) {
            filematch = 0;                                                 //  reject if any ignore string in record
            break;
         }
         
         if (ignorerule == ignore_rec_all && recignore == nigs) {
            filematch = 0;                                                 //  reject if all ignore strings in record
            break;
         }
      }
      
      if (recmatch > 0) {
         if (matchrule == match_rec_all && recmatch < nsrs) recmatch = 0;  //  ignore record without all match strings
         if (recignore > 0) {
            if (ignorerule == ignore_match_any) recmatch = 0;              //  ignore record with any ignore string 
            if (ignorerule == ignore_match_all)
                if (recignore == nigs) recmatch = 0;                       //  ignore record with all ignore strings 
         }
      }

      if (recmatch > 0) {
         filematch += recmatch;                                            //  find matches for entire file
         for (ii = 0; ii < nsrs; ii++) Fmatch[ii] += Rmatch[ii];
      }
      
      if (recignore > 0)                                                   //  ignore matches for entire file
         for (ii = 0; ii < nigs; ii++) Fignore[ii] += Rignore[ii];
   }

   fclose(fid);

   if (filematch == 0) return 0;                                           //  reject file with no match strings
   
   if (ignorerule == ignore_all && nigs > 0) {
      for (ii = 0; ii < nigs; ii++) if (Fignore[ii] == 0) break;           //  reject files with all ignore strings
      if (ii == nigs) return 0;
   }
   
   if (matchrule == match_all) {
      for (ii = 0; ii < nsrs; ii++) if (Fmatch[ii] == 0) return 0;         //  reject files without all match strings
   }
   
   if (! listmatch) {
      wprintf(mLog," %5d %s \n",filematch,filename);                       //  output match count and file name
      return filematch;
   }

   //  file is selected and detail is wanted, read again to list matching records

   snprintf(buff1,999,"\n %s \n",filename);                                //  output file name in bold
   wprintx(mLog,0,buff1,boldfont);

   fid = fopen(filename,"r");                                              //  open file
   if (! fid) return 0;

   while (true)
   {
      prec = fgets_trim(buff1,999,fid,1);                                  //  read next record
      if (! prec) break;

      strcpy(buff2,buff1);                                                 //  copy before destruction by strtok()

      for (ii = 0; ii < Smax; ii++) Rmatch[ii] = Rignore[ii] = 0;          //  no strings found in record yet
      recmatch = recignore = 0;

      while (true)
      {
         token = strtok(prec,delims[0]);                                   //  get next string defined by delimiters
         if (! token) break;
         prec = 0;

         for (ii = 0; ii < nsrs; ii++)
         {
            if (MatchWild(srstrings[ii],token) == 0) {                     //  match with all search strings
               Rmatch[ii]++;
               if (Rmatch[ii] == 1) recmatch++;                            //  search strings found, 0...nsrs
            }
         }

         for (ii = 0; ii < nigs; ii++)
         {
            if (MatchWild(igstrings[ii],token) == 0) {                     //  match with all ignore strings
               Rignore[ii]++;
               if (Rignore[ii] == 1) recignore++;                          //  ignore strings found, 0...nigs
            }
         }
      }
      
      if (recmatch > 0) {
         if (matchrule == match_rec_all && recmatch < nsrs) recmatch = 0;  //  ignore record without all match strings
         if (recignore > 0) {
            if (ignorerule == ignore_match_any) recmatch = 0;              //  ignore record with any ignore string 
            if (ignorerule == ignore_match_all)
                if (recignore == nigs) recmatch = 0;                       //  ignore record with all ignore strings 
         }
      }

      if (recmatch) {
         filematch += recmatch;                                            //  matches for entire file
         for (ii = 0; ii < nsrs; ii++) Fmatch[ii] += Rmatch[ii];
         for (ii = 0; ii < nigs; ii++) Fignore[ii] += Rignore[ii];
      }

      if (recmatch) wprintf(mLog,"     %s \n",buff2);
   }

   fclose(fid);
   return filematch;
}



